import 'package:flutter/material.dart';
import 'package:path_stack/path_stack.dart';

import 'advanced_tab_example_pages.dart';

class AppPaths {
  static const String tabsCategory = "tabs/";
}

/// A double-nested page stack that also supports full-screen paths and back history support.

/// Has a main tab menu that wraps the main (persistent) routes.
/// Settings contains a secondary Tab Menu that also retains its state.
/// Also supports 2 fullscreen routes (Compose and DetailsPage) that are NOT  stateful.
/// with DetailsPage showing an example of route args being parsed (/details/83 -> DetailsPage(itemId: 83))
/// To open DetailsView, click any of the list items.
class AdvancedTabExample extends StatefulWidget {
  @override
  _AdvancedTabExampleState createState() => _AdvancedTabExampleState();
}

ValueNotifier<String> pathNotifier = ValueNotifier(AppPaths.tabsCategory + HomePage.path);
GlobalKey<PathStackState> pathStackKey = GlobalKey();

/// Shows how you could implement a simple history stack by hand (allowing web-like back behavior)
List<String> history = [];

// Move the history back 1 entry
void goBack() {
  if (history.length > 1) {
    String prevPath = history[history.length - 2];
    history..removeLast()..removeLast(); // remove last 2 history entries
    pathNotifier.value = prevPath; // switch, adding a new history entry
  }
}

// Go back in history, and grab the first route that does not match the provided route.
// Used in this demo to close all of the details pages you might have in your history stack
void popMatching(String value) {
  int index = history.lastIndexWhere((element) => element.contains(DetailsPage.path) == false);
  List<String> newHistory = List.from(history)..removeRange(index + 1, history.length);
  pathNotifier.value = newHistory.last;
  history = newHistory;
}

class _AdvancedTabExampleState extends State<AdvancedTabExample> {
  @override
  Widget build(BuildContext context) {
    return ValueListenableBuilder<String>(
      valueListenable: pathNotifier,
      builder: (_, currentPath, __) {
        /// Add entry to history each time we change path
        if (history.isEmpty || currentPath != history.last) {
          history.add(currentPath);
        }

        /// Root level path-stack, contains our main tabbed scaffold, and any full-screen routes
        return PathStack(
          path: currentPath,
          key: pathStackKey,
          routes: {
            [AppPaths.tabsCategory]: PathStack(
              // Main scaffold is wrapped here
              scaffoldBuilder: (_, stack) => MainScaffold(child: stack),
              transitionBuilder: (_, stack, animation) => FadeTransition(opacity: animation, child: stack),
              routes: {
                // Home
                [HomePage.path]: HomePage("").toStackRoute(),
                // Settings
                [SettingsHome.path]: PathStack(
                  // Settings scaffold is wrapped here
                  scaffoldBuilder: (_, child) => SettingsScaffold(child: child),
                  routes: {
                    [ProfileSettings.path]: ProfileSettings("").toStackRoute(),
                    [AlertSettings.path]: AlertSettings("").toStackRoute(),
                    [BillingSettings.path]: BillingSettings("").toStackRoute(maintainState: false),
                  },
                ).toStackRoute(),
              },
            ).toStackRoute(),

            /// Full-screen Route
            [ComposePage.path]: ComposePage().toStackRoute(maintainState: false),
            [DetailsPage.path + ":id"]: StackRouteBuilder(
              maintainState: false,
              builder: (_, args) => DetailsPage(itemId: args["id"]),
            )
          },
        );
      },
    );
  }
}

class BackBtn extends StatelessWidget {
  Widget build(BuildContext context) => OutlinedButton(onPressed: goBack, child: Text("<< Back"));
}
