import 'dart:async';

import 'package:flutter/material.dart';
import 'package:path_stack/src/path_stack_path_provider.dart';
import 'package:path_to_regexp/path_to_regexp.dart';
import 'stack_route_builder.dart';
export 'stack_route_builder.dart';
export 'path_stack_path_provider.dart';

class PathStack extends StatefulWidget {
  static const kDefaultDuration = Duration(milliseconds: 250);

  final String? path;
  final String? basePath;
  final Map<List<String>, StackRouteBuilder> routes;
  final bool caseSensitive;
  final Widget Function(BuildContext context, Widget child)? scaffoldBuilder;
  final Widget Function(BuildContext context, Widget child, AnimationController animation)? transitionBuilder;
  final Widget Function(BuildContext context)? unknownPathBuilder;
  final void Function(String path, Map<String, String> args, {String matchedRoute})? onRouteChanging;
  final Duration transitionDuration;

  PathStack(
      {Key? key,
      this.path,
      required this.routes,
      this.basePath,
      this.caseSensitive = false,
      this.scaffoldBuilder,
      this.transitionBuilder,
      this.transitionDuration = kDefaultDuration,
      this.unknownPathBuilder,
      this.onRouteChanging})
      : super(key: key) {}

  static PathStackState? of(BuildContext context) {
    _InheritedPathStack? stack = context.dependOnInheritedWidgetOfExactType<_InheritedPathStack>();
    return stack?.state;
  }

  @override
  PathStackState createState() => PathStackState();
}

class PathStackState extends State<PathStack> with SingleTickerProviderStateMixin {
  // A cache of widgets, by path. Each widget is rebuilt each time it's shown, but the state is retained.
  Map<String, Widget> _knownRoutes = {};
  String? _matchedRoutePath;
  late AnimationController transitionInAnim = AnimationController(vsync: this, duration: widget.transitionDuration)
    ..forward();

  Map<String, String> _args = {};
  Map<String, String> get args => _args;

  String get basePath => widget.basePath ?? PathStack.of(context)?.currentRoutingPath ?? "";
  // Tries and finds a path value. Respect widget.path first, then look for a parent PathStack, then look for a PathStackPathProvider
  String get path => (widget.path ?? PathStack.of(context)?.path ?? PathStackPathProvider.of(context)?.path ?? "");

  String get currentRoutingPath =>
      (PathStack.of(context)?.currentRoutingPath ?? "") + (_matchedRoutePath?.split(",")[0] ?? "");

  @override
  Widget build(BuildContext context) {
    // Try and find a known route for the current path
    _RoutePathWithBuilder routeToShow = getRouteBuilderForPath(path);
    // Check onBeforeChange, if  this returns false we still need to finish this method, but we should make no changes.
    bool allowChange = routeToShow.content.onBeforeEnter?.call(routeToShow.path) ?? true;
    print("currentRoutingPath: $currentRoutingPath");

    if (allowChange) {
      // Extract queryString and path args
      _args = getArgsFromPath(route: basePath + routeToShow.path, path: path);

      // Check that we have actually changed
      if (_matchedRoutePath != routeToShow.path) {
        // Inform listeners we're about to change
        widget.onRouteChanging?.call(path, args, matchedRoute: routeToShow.path);
        // Use notification to allow any ancestors listen
        PathStackRouteChangingNotification(path, args, matchedRoute: routeToShow.path).dispatch(context);
        // Trigger transitionIn anim
        transitionInAnim.forward(from: 0);
        _matchedRoutePath = routeToShow.path;
      }

      // build the route, add to knownRoutes
      _knownRoutes[routeToShow.path] = routeToShow.build(context, args);
    }

    /// Create IndexedStack
    // Get a list of the children from our known cache
    List<Widget> children = _knownRoutes.values.toList(); // Make a copy of the current knownChildren
    // Get the index for the current child, based on string path
    int index = _knownRoutes.keys.toList().indexWhere((r) => r == _matchedRoutePath);
    Widget content = IndexedStack(index: index, children: children);
    if (routeToShow.content.maintainState == false) removeRouteNextFrame(routeToShow);

    /// Wrap builders
    content = widget.transitionBuilder?.call(context, content, transitionInAnim) ?? content;
    content = widget.scaffoldBuilder?.call(context, content) ?? content;

    /// Wrap content with inherited widget so we can look it up
    return _InheritedPathStack(child: content, state: this);
  }

  void clearState() => _knownRoutes.clear();

  @override
  void dispose() {
    clearState();
    super.dispose();
  }

  bool checkPathMatchesRoute(String path, List<String> routes) {
    //bool allowSuffix(String value) => value.endsWith("/");
    // The first entry in the list of routes will determine whether it allowsSuffix or not.
    // We'll make a special exception or "/" as I cant think of a use case for it and it can cause a lot of confusion... TODO: revisit this decision after more use-case testing
    bool allowSuffix = routes[0].endsWith("/") && routes[0] != "/";
    for (var i = 0; i < routes.length; i++) {
      //print("checking: $basePath${routes[i]}");
      final regExp = pathToRegExp('$basePath${routes[i]}', prefix: allowSuffix, caseSensitive: widget.caseSensitive);
      if (regExp.hasMatch("$path")) return true;
    }
    return false;
  }

  _RoutePathWithBuilder getRouteBuilderForPath(String path) {
    _RoutePathWithBuilder? matchingRoute;
    // For each route, check all paths it has defined and see if we have a match
    widget.routes.forEach((key, value) {
      if (checkPathMatchesRoute(path, key) && matchingRoute == null) {
        matchingRoute = _RoutePathWithBuilder(key.join(","), value);
      }
    });
    // If unable to find route, fallback to unknownPathBuilder or a default 404 page.
    if (matchingRoute == null) {
      print("WARNING: Unable to find a route for $path");
      matchingRoute = _RoutePathWithBuilder("404", StackRouteBuilder(builder: (c, _) {
        return widget.unknownPathBuilder?.call(c) ??
            Scaffold(
              body: Center(
                child: Column(
                  mainAxisSize: MainAxisSize.min,
                  children: [
                    Text("Page Not Found :(", style: TextStyle(fontSize: 26)),
                    Text("Unable to match a route to: '$path'"),
                  ],
                ),
              ),
            );
      }));
    } else {
      //print("RouteMatched, ${matchingRoute?.path}");
    }
    return matchingRoute!;
  }

  Map<String, String> getArgsFromPath({required String route, required String path}) {
    // Get path params
    List<String> paramList = [];
    final regExp = pathToRegExp(route, parameters: paramList);
    Map<String, String> pathMap = {};
    if (regExp.hasMatch(path)) {
      final match = regExp.matchAsPrefix(path)!;
      pathMap = extract(paramList, match);
    }
    // Get queryString params
    Map<String, String> queryMap = Uri.parse(path).queryParameters;
    // Combine path and query args into one, the query args will override the path items with the same id
    final o = {...pathMap, ...queryMap};
    return o;
  }

  void removeRouteNextFrame(_RoutePathWithBuilder routeToShow) =>
      scheduleMicrotask(() => _knownRoutes.remove(routeToShow.path));
}

class _RoutePathWithBuilder {
  _RoutePathWithBuilder(this.path, this.content);
  final String path;
  final StackRouteBuilder content;

  Widget build(BuildContext context, Map<String, String> args) {
    return _KeyedWidget(
      key: ValueKey(path),
      child: content.builder.call(context, args),
    );
  }
}

class _KeyedWidget extends StatelessWidget {
  const _KeyedWidget({Key? key, required this.child}) : super(key: key);
  final Widget child;
  @override
  Widget build(BuildContext context) => child;
}

class _InheritedPathStack extends InheritedWidget {
  _InheritedPathStack({Key? key, required Widget child, required this.state}) : super(key: key, child: child);
  final PathStackState state;
  @override
  bool updateShouldNotify(covariant InheritedWidget oldWidget) => true;
}

class PathStackRouteChangingNotification extends Notification {
  PathStackRouteChangingNotification(this.path, this.args, {this.matchedRoute});
  final String path;
  final String? matchedRoute;
  final Map<String, String> args;
}
